/*
 * Generate a VGA signal using a PIC32 microcontroller.
 *
 * Copyright (C) 2017, 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "debug.h"
#include "display.h"
#include "font.h"
#include "image.h"
#include "init.h"
#include "pic32_c.h"
#include "utils.h"
#include "viewport.h"

/* Specific functionality. */

#include "devconfig.h"
#include "main.h"
#include "vga_display.h"



/* Define the scroll increment depending on how many DMA channels are used.
   CPU-driven transfers permit an increment of 1. */

#ifdef TRANSFER_CPU
#define SCROLL_XSTEP    1
#else
#define SCROLL_XSTEP    LINE_CHANNELS
#endif



/* Define timers if not indicated in the build configuration. */

#ifndef LINE_TIMER
#define LINE_TIMER      2
#endif

#ifndef TRANSFER_TIMER
#define TRANSFER_TIMER  0
#endif

/* Define different output ports and pins for parallel mode. */

#ifndef PARALLEL_MODE
#define VGA_OUTPUT      PORTB
#define OC1_PIN         RPA0R
#define OC2_PIN         RPA1R
#define LED_PIN         (1 << 3)
#else
#define VGA_OUTPUT      PM_REG(0, PMxDIN)
#define OC1_PIN         RPB4R
#define OC2_PIN         RPB5R
#define LED_PIN         (1 << 2)
#endif



/* Initialise memory for a multi-frame display. */

static Display(banner_config, LINE_LENGTH, 10, 1);
static Display(display_config, LINE_LENGTH, LINE_COUNT - 10, FRAME_COUNT);



/* Bundled image and font data. */

extern image_t screendata;
extern image_t sprite;

static SpriteOverwriting(scr, &screendata, &display_config, FRAME_COUNT, -1);

extern uint8_t fontchars[];
extern font_range_t fontbase;

static Font(font_config, fontchars, &fontbase);



/* Plot the revealed region at the edge of the screen after scrolling. */

static void plot_screen_edge(int xorigin, int yorigin, int xstep, int ystep)
{
    image_update_scrolled_tiled_image(&scr, xorigin, yorigin, xstep, ystep);
}



/* Update the sprite position. */

static void update_sprite_position(sprite_t *s, int *xdir, int *ydir)
{
    position_t *last = image_get_stored_position(s, wrap_value(display_config.frame - 1, display_config.frames));
    position_t *next = image_get_sprite_position(s);

    if (((*xdir > 0) && (last->x < display_config.line_length - s->image->width - *xdir)) ||
        ((*xdir < 0) && (last->x > -*xdir)))
    {
        next->x = last->x + *xdir;
    }
    else
    {
        next->x = last->x;
        if (*xdir)
            *xdir = -*xdir;
    }

    if (((*ydir > 0) && (last->y < display_config.line_count - s->image->height - *ydir)) ||
        ((*ydir < 0) && (last->y > -*ydir)))
    {
        next->y = last->y + *ydir;
    }
    else
    {
        next->y = last->y;
        if (*ydir)
            *ydir = -*ydir;
    }
}

/* Move a sprite around on the framebuffer. */

static void animate(uint32_t delay)
{
    /* Stores of background details, replotted when moving the sprite. */

    Sprite(s1, &sprite, &display_config, FRAME_COUNT, 0x0c);
    Sprite(s2, &sprite, &display_config, FRAME_COUNT, 0x0c);
    Sprite(s3, &sprite, &display_config, FRAME_COUNT, 0x0c);

    /* Scrolling viewport. */

    Viewport(v, &display_config, SCROLL_XSTEP, 1, plot_screen_edge);

    /* Reset the viewport to a well-defined state. */

    viewport_set_origin(&v, 0, 0);

    /* Sprite and scrolling directions. */

    int xdir[] = {1, -1, 0}, ydir[] = {1, 0, -1};

    /* Sprite positions. */

    image_set_sprite_position(&s1, 0, 0);
    image_set_sprite_position(&s2, display_config.line_length - s2.image->width, 0);
    image_set_sprite_position(&s3, 0, display_config.line_count - s3.image->height);

    /* Animation loop. */

    while (1)
    {
        image_plot_sprite(&s1);
        image_plot_sprite(&s2);
        image_plot_sprite(&s3);

        /* Update the display with the frame details. */

        vga_set_frame(&display_config, 1);
        wait(delay);

        /* Select the next frame to plot to. */

        display_select_next_frame(&display_config);

        /* Prepare the frame for updates. */

        viewport_unplot_sprite_from_image(&v, &s1, &scr);
        viewport_unplot_sprite_from_image(&v, &s2, &scr);
        viewport_unplot_sprite_from_image(&v, &s3, &scr);

        /* Update the origin to scroll the screen. */

        /* Due to the effect of a simple screen start increment in the
           dual channel configuration, horizontal scrolling involves two
           pixel increments and thus requires a two-pixel column to be
           plotted per scrolling increment. */

        /* For vertically-scaled backgrounds, the full resolution image
           is traversed by multiples of the scrolling increment. */

        viewport_update_origin(&v, xdir[0] * SCROLL_XSTEP,
                                   ydir[0]);

        /* Update the position of the sprite being plotted. */

        update_sprite_position(&s1, &xdir[0], &ydir[0]);
        update_sprite_position(&s2, &xdir[1], &ydir[1]);
        update_sprite_position(&s3, &xdir[2], &ydir[2]);
    }
}

/* Set up a banner. */

static void setup_banner(void)
{
    int frame, x = 0, y = 0;

    for (frame = 0; frame < banner_config.frames; frame++)
    {
        /* Obtain the frame. */

        display_select_frame(&banner_config, frame);

        write_string(&banner_config, &font_config, "VGA display example", &x, &y, 0xff);
    }

    display_select_frame(&banner_config, 0);
}

/* Set up a background. */

static void setup_background(void)
{
    int frame;

    for (frame = 0; frame < display_config.frames; frame++)
    {
        /* Obtain the frame. */

        display_select_frame(&display_config, frame);

        /* Tile the image on the screen. */

        image_tile_sprite(&scr, 0, 0, display_config.line_length,
                                      display_config.line_count,
                                      0, 0);
    }

    display_select_frame(&display_config, 0);
}

static void setup(void)
{
    setup_banner();
    setup_background();
}



/* Main program. */

void main(void)
{
    init_memory();
    init_pins();
    init_outputs();

    unlock_config();
    config_oc();
    config_uart();
    lock_config();

#ifndef TRANSFER_CPU
    init_dma();
#endif

#ifdef PARALLEL_MODE
    init_pm();

    /* Configure parallel master mode. */

    pm_init(0, 0b10);
    pm_set_output(0, 1, 0);
    pm_on(0);
#endif

    uart_init(1, FPB, 115200);
    uart_on(1);

    /* Set the display to a well-defined state. */

    init_display(&banner_config);
    init_display(&display_config);

    /* Initialise VGA output with one or two line channels, configuring a line
       timer and any transfer timer, with an initiating channel being introduced
       if a transfer timer is specified. */

    init_vga_with_timers(&banner_config, LINE_CHANNELS, LINE_TIMER, TRANSFER_TIMER);
    vga_add_window(&display_config);

    /* Configure VGA output transfer to the output register, also configuring
       output compare units for horizontal and vertical sync. */

    vga_configure_transfer(VGA_OUTPUT);
    vga_configure_sync(1, 2);

    interrupts_on();

    /* Move a sprite around on the screen with a delay between each movement. */

    setup();
    animate(1);
}



/* Exception and interrupt handlers. */

void exception_handler(void)
{
    blink(1 << 22, PORTA, LED_PIN);
}

void interrupt_handler(void)
{
    vga_interrupt_handler();
}



/* Peripheral pin configuration. */

void config_oc(void)
{
    /* Map OC1. */

    REG(OC1_PIN) = 0b0101;

    /* Map OC2. */

    REG(OC2_PIN) = 0b0101;
}

void config_uart(void)
{
    /* Map U1RX to RPB13. */

    REG(U1RXR) = 0b0011;        /* U1RXR<3:0> = 0011 (RPB13) */

    /* Map U1TX to RPB15. */

    REG(RPB15R) = 0b0001;       /* RPB15R<3:0> = 0001 (U1TX) */

    /* Set RPB13 to input. */

    SET_REG(TRISB, 1 << 13);
}
