/*
 * Some simple debugging functions.
 *
 * Copyright (C) 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "pic32_c.h"
#include "debug.h"
#include "utils.h"



/* Register output functions using UART1. */

void rbits(uint32_t reg)
{
    bits(REG(reg), 4);
}

void rhex(uint32_t reg)
{
    hex(REG(reg), 4);
}

/* Value output functions using UART1. */

void bits(uint32_t val, int bytes)
{
    uint32_t mask;

    for (mask = (1 << 31); mask; mask >>= 1)
        if (val & mask)
            uart_write('1');
        else
            uart_write('0');
}

void hex(uint32_t val, int bytes)
{
    uint32_t mask;
    uint8_t digit, shift, start = bytes * 8 - 4;

    for (mask = (0b1111 << start), shift = start; mask; mask >>= 4, shift -= 4)
    {
        digit = (val & mask) >> shift;
        if (digit > 9)
            uart_write('A' + digit - 10);
        else
            uart_write('0' + digit);
    }
}



/* General input/output functions. */

int uart_can_read(int uart)
{
    return REG(UART_REG(uart, UxSTA)) & 1;
}

char uart_read_char(int uart)
{
    return (char) REG(UART_REG(uart, UxRXREG));
}

int uart_can_write(int uart)
{
    return !(REG(UART_REG(uart, UxSTA)) & (1 << 9)); /* UTXBF (buffer full) */
}

void uart_write_char(int uart, char c)
{
    while (!uart_can_write(uart)); /* busy loop */

    REG(UART_REG(uart, UxTXREG)) = c;
}



/* Functions using UART1. */

void uart_write(char c)
{
    uart_write_char(1, c);
}

void uart_write_nl(void)
{
    uart_write('\r');
    uart_write('\n');
}

void uart_write_string(const char *s)
{
    while (*s)
        uart_write(*s++);
}



/* Blink an attached LED with delays implemented using a loop. */

void blink(uint32_t delay, uint32_t port, uint32_t pins)
{
    /* Clear outputs (LED). */

    CLR_REG(port, pins);

    while (1)
    {
        wait(delay);

        /* Invert outputs (LED). */

        INV_REG(port, pins);
    }
}
