/*
 * Common font-related functions.
 *
 * Copyright (C) 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdlib.h>

#include "display.h"
#include "font.h"



/* Space is a special case. */

static char_definition_t space = {
    .width = 3, .height = 9, .data = {0, 0, 0, 0, 0, 0, 0, 0, 0}
    };



/* Return the character definition for the given character. */

char_definition_t *get_char_definition(font_config_t *fcfg, char c)
{
    /* Obtain the offset from the table and combine it with the character
       definitions base address. */

    if ((c >= fcfg->range->base) && (c < fcfg->range->limit))
        return (char_definition_t *) ((uint8_t *) fcfg->chars + fcfg->range->table[c - fcfg->range->base]);
    else
        return &space;
}

/* Write a character to the screen, returning the next character position. */

int write_char(display_config_t *cfg, font_config_t *fcfg, char c,
               int x, int y, uint8_t colour)
{
    int sy, dx, dy;
    char_definition_t *chardef = get_char_definition(fcfg, c);
    uint8_t *imagerow = chardef->data,
            *displayline = cfg->screen_start + y * cfg->line_length,
            bitmap, mask;

    for (sy = 0, dy = y; (sy < chardef->height) && (dy < cfg->line_count); sy++, dy++)
    {
        if (displayline >= cfg->screen_limit)
            displayline -= cfg->screen_size;

        bitmap = *imagerow;

        for (dx = x, mask = 1 << (chardef->width - 1); (mask) && (dx < cfg->line_length); dx++, mask >>= 1)
        {
            if (bitmap & mask)
                displayline[display_get_position(cfg, dx)] = colour;
        }

        imagerow += 1;
        displayline += cfg->line_length;
    }

    return x + chardef->width;
}

/* Write a string to the screen, updating the coordinates to indicate the next
   character position. */

void write_string(display_config_t *cfg, font_config_t *fcfg, const char *s,
                  int *x, int *y, uint8_t colour)
{
    const int line_height = 9;

    while (*s)
    {
        if (*x + get_char_definition(fcfg, *s)->width > cfg->line_length)
        {
            *x = 0; *y += line_height;
        }

        if (*y + line_height >= cfg->line_count)
            break;

        *x = write_char(cfg, fcfg, *s, *x, *y, colour);
        s++;
    }
}
